import base64
from typing import cast
import uuid

from fastapi import HTTPException, Response
from fastapi.datastructures import QueryParams
from sqlmodel import select

from openec2.libvirt import LibvirtSingleton
from openec2.config import OpenEC2Config, ConfigSingleton
from openec2.utils.qemu import create_cow_copy
from openec2.db import DatabaseDep
from openec2.db.instance import Instance
from openec2.db.image import AMI
from openec2.db.user import User
from openec2.api.run_instances import RunInstanceResponse, RunInstanceInstanceSet
from openec2.api.describe_instances import describe_instance
from openec2.utils.array import parse_array_objects
from openec2.ipam import (
    get_available_ipv4,
    is_ipv4_available,
    add_instance_dhcp_mapping,
)
from openec2.utils.ip import generate_available_mac


def create_libvirt_domain(
    name: str,
    memory: int,
    vcpu: int,
    ami_path: str,
    mac: str,
) -> str:
    extra_devices = ""
    if ConfigSingleton.of().config.debug:
        extra_devices += """
        <video>
            <model type='qxl' />
        </video>
        <graphics type="spice" port="-1" tlsPort="-1" autoport="yes">
            <image compression="off"/>
        </graphics>
"""

    return f"""
<domain type='kvm'>
    <name>{name}</name>
    <memory unit='MiB'>{memory}</memory>
    <vcpu placement='static'>{vcpu}</vcpu>
    <os>
        <type arch='x86_64'>hvm</type>
        <boot dev='hd' />
        <smbios mode='sysinfo' />
    </os>
    <sysinfo type='smbios'>
        <system>
            <entry name='serial'>ds=nocloud;s=http://192.168.122.1:8000/private/cloudinit/{name}/</entry>
        </system>
    </sysinfo>
    <features>
        <acpi />
        <apic />
        <vmport state='off' />
    </features>
    <clock offset='utc'>
        <timer name='rtc' tickpolicy='catchup'/>
        <timer name='pit' tickpolicy='delay'/>
        <timer name='hpet' present='no'/>
    </clock>
    <pm>
        <suspend-to-mem enabled='no'/>
        <suspend-to-disk enabled='no'/>
    </pm>
    <devices>
        {extra_devices}
        <disk type='file' device='disk'>
            <driver name='qemu' type='qcow2'/>
            <source file='{ami_path}'/>
            <target dev='vda' bus='virtio'/>
        </disk>
        <rng model="virtio">
            <backend model="random">/dev/urandom</backend>
        </rng>
        <interface type="network">
            <source network="default"/>
            <mac address="{mac}"/>
            <model type="virtio"/>
        </interface>
    </devices>
</domain>
"""


def run_instances(
    params: QueryParams,
    config: OpenEC2Config,
    db: DatabaseDep,
    user: User,
):
    image_id = params["ImageId"]
    instance_type = params["InstanceType"]

    instance_type = config.instances.types.get(params["InstanceType"])
    if instance_type is None:
        raise Exception(f"Unknown instance type {params['InstanceType']}")

    ami = db.exec(select(AMI).where(AMI.id == image_id)).first()
    if ami is None:
        raise Exception(f"Unknown AMI {image_id}")

    if ami.deregistered:
        raise HTTPException(
            status_code=400,
            detail="AMI is deregistered and cannot be used anymore",
        )

    # Parse tags
    # TODO: broken
    tags: dict[str, str] = {}
    for spec in parse_array_objects("TagSpecification", cast(dict, params)):
        for raw_tag in parse_array_objects("Tag", spec):
            tags[raw_tag["Key"]] = raw_tag["Value"]

    # Get a private IPv4
    instance_id = uuid.uuid4().hex
    private_ipv4 = params.get(
        "PrivateIpAddress",
        get_available_ipv4(db),
    )
    if not is_ipv4_available(private_ipv4, db):
        raise HTTPException(
            status_code=400,
            detail="Used IPv4",
        )
    mac = generate_available_mac(db)
    add_instance_dhcp_mapping(instance_id, mac, private_ipv4, db)

    # Prepare the instance directory
    config.instances.location.mkdir(exist_ok=True)
    disk = config.instances.location / instance_id
    create_cow_copy(
        config.images / ami.id,
        disk,
        f"{instance_type.disk}G",
    )

    instance = Instance(
        id=instance_id,
        imageId=image_id,
        tags=tags,
        userData=base64.b64decode(value).decode()
        if (value := params.get("UserData")) is not None
        else None,
        privateIPv4=private_ipv4,
        interfaceMac=mac,
        owner_id=user.id,
    )
    db.add(instance)
    print("Inserted new instance")

    conn = LibvirtSingleton.of().connection
    domain = conn.defineXML(
        create_libvirt_domain(
            instance_id,
            instance_type.memory,
            int(instance_type.vcpu),
            str(config.instances.location / instance_id),
            mac,
        ),
    )
    domain.create()
    description = describe_instance(instance, domain)

    db.commit()
    return Response(
        RunInstanceResponse(
            request_id=uuid.uuid4().hex,
            instance_set=RunInstanceInstanceSet(
                item=[description],
            ),
        ).to_xml(),
        media_type="application/xml",
    )
