import uuid

from fastapi import HTTPException, Response
from fastapi.datastructures import QueryParams
from sqlmodel import select

from openec2.libvirt import LibvirtSingleton
from openec2.config import OpenEC2Config
from openec2.db import DatabaseDep
from openec2.db.instance import Instance
from openec2.db.user import User
from openec2.utils.array import parse_array_plain
from openec2.api.shared import InstanceInfo, InstancesSet
from openec2.api.describe_instances import describe_instance_state
from openec2.api.start_instances import StartInstancesResponse


def start_instances(
    params: QueryParams,
    config: OpenEC2Config,
    db: DatabaseDep,
    user: User,
):
    conn = LibvirtSingleton.of().connection
    instances: list[InstanceInfo] = []
    for instance_id in parse_array_plain("InstanceId", params):
        instance = db.exec(select(Instance).where(Instance.id == instance_id)).first()
        if instance is None:
            raise HTTPException(status_code=404, detail="Unknown instance")

        # Check for permission issues
        if instance.owner_id != user.id:
            # TODO: Add the error to the response
            continue

        dom = conn.lookupByName(instance_id)
        prev_state = describe_instance_state(dom)
        if not dom.isActive():
            dom.create()
        current_state = describe_instance_state(dom)

        instances.append(
            InstanceInfo(
                instanceId=instance_id,
                previousState=prev_state,
                currentState=current_state,
            ),
        )

    return Response(
        StartInstancesResponse(
            request_id=uuid.uuid4().hex,
            instancesSet=InstancesSet(
                item=instances,
            ),
        ).to_xml(),
        media_type="application/xml",
    )
