from typing import Annotated, cast
from dataclasses import dataclass
import datetime
from hashlib import sha256
from urllib.parse import quote, parse_qs

from sqlmodel import select
from fastapi import Request, HTTPException, Depends
from fastapi.datastructures import URL, Headers
from cryptography.hazmat.primitives import hashes, hmac

from openec2.config import ConfigSingleton
from openec2.db import DatabaseDep
from openec2.db.user import User


def _hmac_sha256(key: bytes, payload: bytes) -> bytes:
    h = hmac.HMAC(key, hashes.SHA256())
    h.update(payload)
    return h.finalize()


@dataclass
class AWSRequest:
    # The entire used URL
    url: URL

    # Request method
    method: str

    # Query params
    params: dict[str, str]

    # Request headers
    headers: Headers

    # The payload, if we used a POST/PUT
    payload: str | None

    def sign(
        self, secret_access_key: str, region: str, product: str, credential_scope: str
    ) -> str:
        dt = datetime.datetime.fromisoformat(self.headers["X-Amz-Date"])

        canonical_header_string_keys = sorted(
            [name for name in self.headers.keys() if include_in_canonical_string(name)]
        )
        canonical_header_string = (
            "\n".join(
                [
                    f"{name.lower()}:{self.headers[name].strip()}"
                    for name in canonical_header_string_keys
                ]
            )
            + "\n"
        )
        signed_headers = ";".join(canonical_header_string_keys)
        hashed_payload = sha256((self.payload or "").encode()).hexdigest()
        canonical_request = "\n".join(
            [
                self.method.upper(),
                self.url.path or "/",
                # canonical_query_string,
                "",
                canonical_header_string,
                signed_headers,
                hashed_payload,
            ]
        )
        print("Canonical request")
        print(canonical_request)

        hashed_canonical_request = sha256(canonical_request.encode()).hexdigest()
        date = dt.strftime("%Y%m%d")
        string_to_sign = "\n".join(
            [
                "AWS4-HMAC-SHA256",
                dt.strftime("%Y%m%dT%H%M%SZ"),
                credential_scope,
                hashed_canonical_request,
            ]
        )

        print("String to sign")
        print(string_to_sign)

        date_key = _hmac_sha256(f"AWS4{secret_access_key}".encode(), date.encode())
        date_region_key = _hmac_sha256(date_key, region.encode())
        date_region_service_key = _hmac_sha256(date_region_key, product.encode())
        signing_key = _hmac_sha256(date_region_service_key, "aws4_request".encode())
        return _hmac_sha256(signing_key, string_to_sign.encode()).hex()


@dataclass
class AWSAuthentication:
    x_amz_algorithm: str

    x_amz_credential: str

    x_amz_signature: str


def include_in_canonical_string(name: str) -> bool:
    lower = name.lower()
    return lower in ("host", "content-type") or lower.startswith("x-amz")


def get_authentication_info(request: Request) -> AWSAuthentication:
    if request.method == "POST":
        algorithm, rest = request.headers["Authorization"].split(" ", 1)
        auth = {}
        for part in rest.split(","):
            key, value = part.split("=", 1)
            auth[key.strip()] = value.strip()

        return AWSAuthentication(
            x_amz_algorithm=algorithm,
            x_amz_credential=auth["Credential"],
            x_amz_signature=auth["Signature"],
        )

    return AWSAuthentication(
        "",
        "",
        "",
    )


async def check_request_signature(request: Request, db: DatabaseDep):
    # Do not check if we don't care
    if ConfigSingleton.of().config.insecure:
        return

    body = (await request.body()).decode()
    query_params = (
        cast(dict, parse_qs(body))
        if request.method == "POST"
        else cast(dict, request.query_params)
    )

    auth_info = get_authentication_info(request)

    if auth_info.x_amz_algorithm != "AWS4-HMAC-SHA256":
        raise HTTPException(
            status_code=400,
            detail=f"Invalid signature algorithm: {auth_info.x_amz_algorithm}",
        )

    x_amz_credential = auth_info.x_amz_credential

    reversed_parts = x_amz_credential[::-1].split("/", 4)
    access_key, date, region, service, key = [s[::-1] for s in reversed_parts][::-1]
    user = db.exec(select(User).where(User.access_key == access_key)).first()
    if user is None:
        raise HTTPException(status_code=403)

    x_amz_signature = auth_info.x_amz_signature
    signature = AWSRequest(
        url=request.url,
        method=request.method,
        params=query_params,
        headers=request.headers,
        payload=body,
    ).sign(
        user.secret_access_key,
        region,
        service,
        "/".join([date, region, service, key]),
    )

    print(x_amz_signature, signature)
    if x_amz_signature != signature:
        raise HTTPException(status_code=401)

    return user


AWSSignature = Annotated[User, Depends(check_request_signature)]
